import pandas as pd
import collections
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.integrations.handlers.utilities.query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
)


def flatten(d, parent_key="", sep="_"):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class EventbriteUserTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        if query.limit is not None:
            raise NotImplementedError("Limit is not supported for user info table")

        user_info = self.handler.api.get_current_user()

        # Normalize email field
        if "emails" in user_info and isinstance(user_info["emails"], list):
            user_info["email"] = (
                user_info["emails"][0]["email"] if user_info["emails"] else None
            )
            user_info["email_verified"] = (
                user_info["emails"][0]["verified"] if user_info["emails"] else None
            )
            user_info["email_primary"] = (
                user_info["emails"][0]["primary"] if user_info["emails"] else None
            )
            del user_info["emails"]
        else:
            user_info["email"] = None
            user_info["email_verified"] = None
            user_info["email_primary"] = None

        data = pd.DataFrame([user_info])

        # Select columns based on query
        columns = self.get_columns()
        selected_columns = [
            target.parts[-1]
            for target in query.targets
            if isinstance(target, ast.Identifier)
        ]
        if selected_columns:
            columns = [col for col in columns if col in selected_columns]
        data = data[columns]

        return data

    def get_columns(self):
        return [
            "email",
            "email_verified",
            "email_primary",
            "id",
            "name",
            "first_name",
            "last_name",
            "is_public",
            "image_id",
        ]


class EventbriteOrganizationTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        organization_info = self.handler.api.get_user_organizations()

        # Normalize organization data
        organizations = organization_info.get("organizations", [])
        result = pd.DataFrame(organizations)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]

        return result

    def get_columns(self):
        return [
            "_type",
            "name",
            "vertical",
            "parent_id",
            "locale",
            "created",
            "image_id",
            "id",
        ]


class EventbriteCategoryTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        category_info = self.handler.api.list_categories()
        categories = category_info.get("categories", [])
        result = pd.DataFrame(categories)

        select_statement_parser = SELECTQueryParser(
            query, "categoryInfoTable", self.get_columns()
        )

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 100

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]

        select_statement_executor = SELECTQueryExecutor(
            result, selected_columns, where_conditions, order_by_conditions
        )

        result = select_statement_executor.execute_query()

        return result.head(total_results)

    def get_columns(self):
        return [
            "resource_uri",
            "id",
            "name",
            "name_localized",
            "short_name",
            "short_name_localized",
        ]


class EventbriteSubcategoryTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        # Dummy function to simulate API call and response
        category_info = self.handler.api.list_subcategories()

        # Normalizing the category data
        categories = category_info.get("subcategories", [])
        result = pd.DataFrame(categories)

        select_statement_parser = SELECTQueryParser(
            query, "subcategoryInfoTable", self.get_columns()
        )

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 100

        # Normalize nested fields
        parent_category = result["parent_category"].apply(pd.Series)
        parent_category.columns = [f"parent_{col}" for col in parent_category.columns]
        result = pd.concat([result, parent_category], axis=1).drop(
            "parent_category", axis=1
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]

        select_statement_executor = SELECTQueryExecutor(
            result, selected_columns, where_conditions, order_by_conditions
        )

        result = select_statement_executor.execute_query()

        return result.head(total_results)

    def get_columns(self):
        return [
            "resource_uri",
            "id",
            "name",
            "name_localized",
            "parent_category",
            "parent_resource_uri",
            "parent_id",
            "parent_name",
            "parent_name_localized",
            "parent_short_name",
            "parent_short_name_localized",
        ]


class EventbriteFormatTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        # Simulate API call and get the response
        format_info = self.handler.api.list_formats()

        # Normalize format data
        formats = format_info.get("formats", [])
        result = pd.DataFrame(formats)

        select_statement_parser = SELECTQueryParser(
            query, "formatInfoTable", self.get_columns()
        )

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 100

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]

        select_statement_executor = SELECTQueryExecutor(
            result, selected_columns, where_conditions, order_by_conditions
        )

        result = select_statement_executor.execute_query()

        return result.head(total_results)

    def get_columns(self):
        return [
            "resource_uri",
            "id",
            "name",
            "name_localized",
            "short_name",
            "short_name_localized",
        ]


class EventbriteEventDetailsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        conditions = extract_comparison_conditions(query.where)
        allowed_keys = set(["event_id"])

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")
            elif op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            elif op != "=":
                raise NotImplementedError(f"Unknown op: {op}")
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params["max_results"] = query.limit.value

        if "event_id" not in params:
            # search not works without searchQuery, use 'London'
            params["event_id"] = "717926867587"

        event_details = self.handler.api.get_event(params["event_id"])

        # Normalize event data
        flat_event_details = flatten(event_details)
        result = pd.DataFrame([flat_event_details])

        for col in ["name", "description", "start", "end", "logo"]:
            if col in result.columns:
                result = pd.concat(
                    [result, result[col].apply(pd.Series).add_prefix(f"{col}_")],
                    axis=1,
                )
                result = result.drop([col], axis=1)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        return [
            "name_text",
            "name_html",
            "description_text",
            "description_html",
            "url",
            "start_timezone",
            "start_local",
            "start_utc",
            "end_timezone",
            "end_local",
            "end_utc",
            "organization_id",
            "created",
            "changed",
            "published",
            "capacity",
            "capacity_is_custom",
            "status",
            "currency",
            "listed",
            "shareable",
            "online_event",
            "tx_time_limit",
            "hide_start_date",
            "hide_end_date",
            "locale",
            "is_locked",
            "privacy_setting",
            "is_series",
            "is_series_parent",
            "inventory_type",
            "is_reserved_seating",
            "show_pick_a_seat",
            "show_seatmap_thumbnail",
            "show_colors_in_seatmap_thumbnail",
            "source",
            "is_free",
            "version",
            "summary",
            "facebook_event_id",
            "logo_id",
            "organizer_id",
            "venue_id",
            "category_id",
            "subcategory_id",
            "format_id",
            "id",
            "resource_uri",
            "is_externally_ticketed",
            "logo_crop_mask",
            "logo_original",
            "logo_id",
            "logo_url",
            "logo_aspect_ratio",
            "logo_edge_color",
            "logo_edge_color_set",
        ]


class EventbriteEventsTable(APITable):
    def __init__(self, handler):
        self.handler = handler

    def select(self, query: ast.Select) -> pd.DataFrame:
        self.handler.connect()

        conditions = extract_comparison_conditions(query.where)
        allowed_keys = set(["organization_id"])

        params = {}
        for op, arg1, arg2 in conditions:
            if op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            else:
                raise NotImplementedError(
                    f"Unsupported operation or field: {op} {arg1}"
                )

        if "organization_id" not in params:
            raise ValueError("Organization ID must be provided")

        event_list = self.handler.api.list_events(params["organization_id"])
        result = pd.DataFrame(event_list["events"])

        # Normalize event data and split nested dictionaries into separate columns
        result = pd.concat(
            [
                result.drop(["name", "description", "start", "end", "logo"], axis=1),
                result["name"].apply(pd.Series).add_prefix("name_"),
                result["description"].apply(pd.Series).add_prefix("description_"),
                result["start"].apply(pd.Series).add_prefix("start_"),
                result["end"].apply(pd.Series).add_prefix("end_"),
                result["logo"].apply(pd.Series).add_prefix("logo_"),
            ],
            axis=1,
        )

        select_statement_parser = SELECTQueryParser(
            query, "subcategoryInfoTable", self.get_columns()
        )

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        total_results = result_limit if result_limit else 100

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]

        select_statement_executor = SELECTQueryExecutor(
            result, selected_columns, where_conditions, order_by_conditions
        )

        result = select_statement_executor.execute_query()

        return result.head(total_results)

    def get_columns(self):
        return [
            "name_text",
            "name_html",
            "description_text",
            "description_html",
            "url",
            "start_timezone",
            "start_local",
            "start_utc",
            "end_timezone",
            "end_local",
            "end_utc",
            "organization_id",
            "created",
            "changed",
            "published",
            "capacity",
            "capacity_is_custom",
            "status",
            "currency",
            "listed",
            "shareable",
            "online_event",
            "tx_time_limit",
            "hide_start_date",
            "hide_end_date",
            "locale",
            "is_locked",
            "privacy_setting",
            "is_series",
            "is_series_parent",
            "inventory_type",
            "is_reserved_seating",
            "show_pick_a_seat",
            "show_seatmap_thumbnail",
            "show_colors_in_seatmap_thumbnail",
            "source",
            "is_free",
            "summary",
            "organizer_id",
            "venue_id",
            "category_id",
            "subcategory_id",
            "format_id",
            "id",
            "resource_uri",
            "is_externally_ticketed",
        ]



from typing import Optional, Dict

import pandas as pd
from nixtlats import TimeGPT

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.utilities.time_series_utils import get_results_from_nixtla_df
# TODO: add E2E tests.


class TimeGPTHandler(BaseMLEngine):
    """
    Integration with the Nixtla TimeGPT models for
    zero-shot time series forecasting.
    """

    name = "timegpt"

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Create the TimeGPT Handler.
        Requires specifying the target column and usual time series arguments. Saves model config for later usage.
        """
        self.generative = True
        time_settings = args.get("timeseries_settings", {})
        using_args = args["using"]

        mode = 'forecasting'
        if args.get('__mdb_sql_task', False) and args['__mdb_sql_task'].lower() in ('forecasting', 'anomalydetection'):
            mode = args['__mdb_sql_task'].lower()

        if mode == 'forecasting':
            assert time_settings["is_timeseries"], "Specify time series settings in your query"

        model_args = {
            'token': get_api_key('TIMEGPT_TOKEN', using_args, self.engine_storage, strict=True),
            "target": target,
            "freq": using_args.get("frequency", None),
            "finetune_steps": using_args.get("finetune_steps", 0),
            "validate_token": using_args.get("validate_token", False),
            "date_features": using_args.get("date_features", False),
            "date_features_to_one_hot": using_args.get("date_features_to_one_hot", True),
            "clean_ex_first": using_args.get("clean_ex_first", True),
            "level": using_args.get("level", [90]),
            "add_history": using_args.get("add_history", False),
            'mode': mode,
        }

        if time_settings:
            model_args["horizon"] = time_settings["horizon"]
            model_args["order_by"] = time_settings["order_by"]
            model_args["group_by"] = time_settings.get("group_by", [])

        if mode == 'anomalydetection':
            model_args["target"] = using_args['target'] if target is None else target
            model_args["horizon"] = using_args.get('horizon', 1)
            model_args["order_by"] = using_args['order_by']
            model_args["group_by"] = using_args.get("group_by", [])
            model_args['add_history'] = True

        assert isinstance(model_args["level"], list), "`level` must be a list of integers"
        assert all([isinstance(l, int) for l in model_args["level"]]), "`level` must be a list of integers"

        self.model_storage.json_set("model_args", model_args)  # persist changes to handler folder

    def predict(self, df, args={}):
        """ Makes forecasts with the TimeGPT API. """
        model_args = self.model_storage.json_get("model_args")
        args = args['predict_params']
        prediction_df = self._transform_to_nixtla_df(df, model_args)
        timegpt = TimeGPT(token=model_args['token'])

        forecast_df = timegpt.forecast(
            prediction_df,

            # TODO: supporting param override when JOINing with a WHERE clause is blocked by mindsdb_sql#285
            h=args.get("horizon", model_args.get("horizon", 1)),
            freq=args.get("freq", model_args["freq"]),  # automatically infers correct frequency if not provided by user
            level=model_args["level"],
            finetune_steps=args.get('finetune_steps', model_args['finetune_steps']),
            validate_token=args.get('validate_token', model_args['validate_token']),
            date_features=args.get('date_features', model_args['date_features']),
            date_features_to_one_hot=args.get('date_features_to_one_hot', model_args['date_features_to_one_hot']),
            clean_ex_first=args.get('clean_ex_first', model_args['clean_ex_first']),

            # anomaly detection
            add_history=args.get('add_history', model_args['add_history'])  # insample bounds and anomaly detection

            # TODO: enable this post-refactor (#6861)
            # X_df=None,  # exogenous variables
        )

        if model_args['mode'] == 'forecasting':
            results_df = forecast_df[['unique_id', 'ds', 'TimeGPT']]
            results_df = get_results_from_nixtla_df(results_df, model_args)
        elif model_args['mode'] == 'anomalydetection':
            forecast_df['ds'] = pd.to_datetime(forecast_df['ds'])
            results_df = forecast_df.merge(prediction_df, how='inner')  # some rows drop because of TimeGPT's cold start
            results_df['anomaly'] = (results_df['y'] > results_df[f'TimeGPT-hi-{model_args["level"][0]}']) | (results_df['y'] < results_df[f'TimeGPT-lo-{model_args["level"][0]}'])

            forecast_df = results_df  # rewrite forecast_df so that we can reuse code below for prediction intervals
            results_df = get_results_from_nixtla_df(results_df, model_args)
            results_df = results_df.rename({'y': f'observed_{model_args["target"]}'}, axis=1)
        else:
            raise Exception(f'Unsupported prediction mode: {model_args["mode"]}')

        # infer date
        ds_col = model_args["order_by"]
        if not pd.api.types.is_datetime64_any_dtype(results_df[ds_col]):
            results_df[ds_col] = pd.to_datetime(results_df[ds_col])

        results_df = results_df.rename({'TimeGPT': model_args['target']}, axis=1)

        # add prediction intervals
        levels = sorted(model_args['level'], reverse=True)
        for i, level in enumerate(levels):
            if i == 0:
                # NOTE: this should be simplified once we refactor the expected time series output within MindsDB
                results_df['confidence'] = level/100  # we report the highest level as the overall confidence
                results_df['lower'] = forecast_df[f'TimeGPT-lo-{level}']
                results_df['upper'] = forecast_df[f'TimeGPT-hi-{level}']
            else:
                results_df[f'lower_{level}'] = forecast_df[f'TimeGPT-lo-{level}']
                results_df[f'upper_{level}'] = forecast_df[f'TimeGPT-hi-{level}']

        return results_df

    def describe(self, attribute=None):
        model_args = self.model_storage.json_get("model_args")

        if attribute == "model":
            df = pd.DataFrame({"frequency": [model_args["freq"] if model_args["freq"] else "automatic"]})
            return df

        elif attribute == "features":
            df = pd.DataFrame({
                "order by": [model_args["order_by"]],
                "target": model_args["target"]
            })
            if model_args["group_by"]:
                df["group by"] = [model_args["group_by"]]
            return df

        elif attribute == 'info':
            outputs = model_args["target"]
            inputs = [model_args["target"], model_args["order_by"]]
            if model_args["group_by"]:
                inputs.append(model_args["group_by"])
            return pd.DataFrame({"output": outputs, "input": [inputs]})

        else:
            tables = ['info', 'features', 'model']
            return pd.DataFrame(tables, columns=['tables'])

    # TODO: consolidate this method with the ones in time_series_utils.py
    @staticmethod
    def _convert_to_iso(df, date_column):
        # whether values in date_column are numeric (Unix timestamp) or string (date)
        if pd.api.types.is_numeric_dtype(df[date_column]):
            unit = ''
            # ascending unit order
            for u in ['ns', 'us', 'ms', 's']:
                mindate = pd.to_datetime(df[date_column].min(), unit=u, origin='unix')
                maxdate = pd.to_datetime(df[date_column].max(), unit=u, origin='unix')
                if mindate > pd.to_datetime('1970-01-01T00:00:00') and maxdate < pd.to_datetime('2050-12-31T23:59:59'):
                    unit = u
            df[date_column] = pd.to_datetime(df[date_column], unit=unit, origin='unix')
        else: 
            df[date_column] = pd.to_datetime(df[date_column])
        df[date_column] = df[date_column].dt.strftime('%Y-%m-%dT%H:%M:%S')  # convert to ISO 8601 format
        df[date_column] = pd.to_datetime(df[date_column])
        return df

    # TODO: consolidate this method with the ones in time_series_utils.py
    def _transform_to_nixtla_df(self, df, settings_dict, exog_vars=[]):
        nixtla_df = df.copy()
        # Transform group columns into single unique_id column
        gby = settings_dict["group_by"]
        if len(gby) > 1:
            for col in gby:
                nixtla_df[col] = nixtla_df[col].astype(str)
            nixtla_df["unique_id"] = nixtla_df[gby].agg("/".join, axis=1)
            group_col = "ignore this"
        elif len(gby) == 1 and gby[0] is not None:
            group_col = settings_dict["group_by"][0]
        else:
            group_col = '__unique_id'
            nixtla_df[group_col] = '1'

        # Rename columns to statsforecast names
        nixtla_df = nixtla_df.rename(
            {settings_dict["target"]: "y", settings_dict["order_by"]: "ds", group_col: "unique_id"}, axis=1
        )

        columns_to_keep = ["unique_id", "ds", "y"] + exog_vars
        nixtla_df = self._convert_to_iso(nixtla_df, "ds")
        nixtla_df = nixtla_df[columns_to_keep].sort_values(by=['unique_id', 'ds'], ascending=True)  # expects ascending
        nixtla_df['y'] = nixtla_df['y'].astype(float)
        return nixtla_df.reset_index(drop=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: db.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import mindsdb.microservices_grpc.db.common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x08\x64\x62.proto\x1a\x0c\x63ommon.proto\"O\n\x0eHandlerContext\x12\x0f\n\x07\x63ontext\x18\x01 \x01(\t\x12\x14\n\x0chandler_type\x18\x02 \x01(\t\x12\x16\n\x0ehandler_params\x18\x03 \x01(\t\"E\n\x12NativeQueryContext\x12 \n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x0f.HandlerContext\x12\r\n\x05query\x18\x02 \x01(\t\"E\n\x12\x42inaryQueryContext\x12 \n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x0f.HandlerContext\x12\r\n\x05query\x18\x02 \x01(\x0c\"A\n\x0e\x43olumnsContext\x12 \n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x0f.HandlerContext\x12\r\n\x05table\x18\x02 \x01(\t2\xdc\x02\n\tDBService\x12-\n\x07\x43onnect\x12\x0f.HandlerContext\x1a\x0f.StatusResponse\"\x00\x12\x35\n\x0f\x43heckConnection\x12\x0f.HandlerContext\x1a\x0f.StatusResponse\"\x00\x12\x30\n\nDisconnect\x12\x0f.HandlerContext\x1a\x0f.StatusResponse\"\x00\x12/\n\x0bNativeQuery\x12\x13.NativeQueryContext\x1a\t.Response\"\x00\x12/\n\x0b\x42inaryQuery\x12\x13.BinaryQueryContext\x1a\t.Response\"\x00\x12)\n\tGetTables\x12\x0f.HandlerContext\x1a\t.Response\"\x00\x12*\n\nGetColumns\x12\x0f.ColumnsContext\x1a\t.Response\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'db_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _HANDLERCONTEXT._serialized_start=26
  _HANDLERCONTEXT._serialized_end=105
  _NATIVEQUERYCONTEXT._serialized_start=107
  _NATIVEQUERYCONTEXT._serialized_end=176
  _BINARYQUERYCONTEXT._serialized_start=178
  _BINARYQUERYCONTEXT._serialized_end=247
  _COLUMNSCONTEXT._serialized_start=249
  _COLUMNSCONTEXT._serialized_end=314
  _DBSERVICE._serialized_start=317
  _DBSERVICE._serialized_end=665
# @@protoc_insertion_point(module_scope)



import os
import json
import pickle
from uuid import uuid4

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx
from mindsdb.integrations.libs.handler_helpers import action_logger

logger = log.getLogger(__name__)


try:
    import grpc
    from mindsdb.microservices_grpc.executor import executor_pb2_grpc
    from mindsdb.microservices_grpc.executor import executor_pb2
except ImportError:
    logger.error("to user microservice mode please install 'pip install mindsdb[grpc]'")


class ExecutorClientGRPC:

    def __init__(self, session, sqlserver):
        self.id = f"executor_{uuid4()}"

        self.sqlserver = sqlserver
        self.session = session
        self.query = None

        # returns
        self.columns = []
        self.params = []
        self.data = None
        self.state_track = None
        self.server_status = None
        self.is_executed = False

        # additional attributes used in handling query process
        self.sql = ""
        self.sql_lower = ""

        host = os.environ.get("MINDSDB_EXECUTOR_SERVICE_HOST", None)
        port = os.environ.get("MINDSDB_EXECUTOR_SERVICE_PORT", None)
        self.channel = grpc.insecure_channel(f"{host}:{port}")
        self.stub = executor_pb2_grpc.ExecutorServiceStub(self.channel)

    @property
    def _context(self):
        logger.info("%s._context: preparing context. id - %s, session - %s", self.__class__.__name__, self.id, self.session)
        _id = self.id
        connection_id = self.sqlserver.connection_id if hasattr(self.sqlserver, "connection_id") else -1
        session_id = self.session.id
        session = json.dumps(self.session.to_json())
        context = json.dumps(ctx.dump())

        return executor_pb2.ExecutorContext(
            id=_id,
            connection_id=connection_id,
            session_id=session_id,
            session=session,
            context=context,
        )

    def __del__(self):
        self.stub.DeleteExecutor(self._context)

    def _update_attrs(self, response: executor_pb2.ExecutorResponse):
        self.columns = pickle.loads(response.columns)
        self.params = pickle.loads(response.params)
        self.data = pickle.loads(response.data)
        self.state_track = response.state_track
        self.is_executed = response.is_executed
        self.session.from_json(json.loads(response.session))
        self.error_code = response.error_code
        self.error_message = response.error_message
        logger.info("%s._update_attrs: got data from service - %s", self.__class__.__name__, self.data)

    @action_logger(logger)
    def stmt_prepare(self, sql):
        params = executor_pb2.ExecutionContext(context=self._context, sql=sql)
        resp = self.stub.StatementPrepare(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def stmt_execute(self, param_values):
        if self.is_executed:
            return

        params = executor_pb2.StatementExecuteContext(context=self._context, param_values=json.dumps(param_values))
        resp = self.stub.StatementExecute(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def query_execute(self, sql):
        params = executor_pb2.ExecutionContext(context=self._context, sql=sql)
        resp = self.stub.QueryExecute(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def binary_query_execute(self, sql):
        sql = pickle.dumps(sql)
        params = executor_pb2.BinaryExecutionContext(context=self._context, sql=sql)
        resp = self.stub.BinaryQueryExecute(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def execute_external(self, sql):
        params = executor_pb2.ExecutionContext(context=self._context, sql=sql)
        resp = self.stub.ExecuteExternal(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def parse(self, sql):
        params = executor_pb2.ExecutionContext(context=self._context, sql=sql)
        resp = self.stub.Parse(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def do_execute(self):
        if self.is_executed:
            return
        resp = self.stub.DoExecute(self._context)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)

    @action_logger(logger)
    def change_default_db(self, new_db):
        params = executor_pb2.DefaultDBContext(context=self._context, new_db=new_db)
        resp = self.stub.ChangeDefaultDB(params)
        if resp.error_message != "":
            raise Exception(resp.error_message)
        self._update_attrs(resp)
